# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""Module for future CDK components.

Components here are planned to move to the CDK.

TODO!: Add GitHub link here before merging.
"""

from __future__ import annotations

from airbyte_cdk.sql.shared.sql_processor import SqlProcessorBase

__all__ = [
    "SqlProcessorBase",
]
